package polyglot.ext.ejava.ast;

import java.util.*;

import polyglot.ast.*;
import polyglot.ext.ejava.types.EJavaContext;
import polyglot.ext.ejava.types.EJavaTypeSystem;
import polyglot.ext.ejava.types.ExtensionType;
import polyglot.ext.ejava.visit.EJavaTypeBuilder;
import polyglot.ext.jl5.ast.JL5ClassDecl_c;
import polyglot.ext.jl5.types.FlagAnnotations;
import polyglot.main.Report;
import polyglot.types.*;
import polyglot.util.CodeWriter;
import polyglot.util.Position;
import polyglot.visit.*;

public class ExtensionDecl_c extends JL5ClassDecl_c implements ExtensionDecl {
    protected boolean isTop, isLocalTop;
    protected Set cases;
    protected Type topForClass;

    public String mangle(Object x) {
        return x.toString().replace(".", "$$$").replace("<", "$$$").replace(">", "$$$");
    }

    public Set cases() { return cases; }
    public void addCase(ExtensionDecl ed) {
        cases.add(ed);
    }
    public void clearCases() { cases = new HashSet(); }

    protected String extensionName;
    public String extensionName() { return extensionName; }

    public ExtensionDecl_c(Position pos, FlagAnnotations flags, String name, TypeNode forClass, List interfaces, ClassBody body, boolean isTop) {
        super(pos, flags, name, forClass, interfaces, body);
        this.extensionName = name;
        if (!isTop)
            this.name = name+"$"+mangle(forClass);
        this.isTop = isTop;
        isLocalTop = false;
        cases = new HashSet();
    }

    public boolean isTop() { return isTop; }
    public boolean isLocalTop() { return isLocalTop; }
    public void promoteToLocalTop() { isLocalTop = true; }

    public TypeNode forClass() { return superClass(); }
    public ExtensionDecl forClass(TypeNode forClass) {
        return (ExtensionDecl)superClass(forClass);
    }

    protected boolean isMethodOfExtension(MethodInstance m) {
        return type().hasMethodImpl(m);
    }


    protected void disambiguateSuperType(AmbiguityRemover ar) throws SemanticException {
        TypeSystem ts = ar.typeSystem();

        if (this.superClass != null) {
            Type t = this.superClass.type();

            if (! t.isCanonical())
                throw new SemanticException("Could not disambiguate super" + " class of " + type + ".", superClass.position());

            /*
             * // Alex: Changed this next condition so that interfaces are ok if (!
             * t.isClass() && ! t.toClass().flags().isInterface()) throw new
             * SemanticException("Super class " + t + " of " + type + " is not a class.",
             * superClass.position());
             */
            if (! t.isClass() && t.toClass().flags().isInterface())
                throw new SemanticException("extended class " + t + " of " + type + " should be a class, but isn't.", superClass.position());

            if (Report.should_report(Report.types, 3))
                Report.report(3, "setting super type of " + this.type + " to " + t);

            this.type.superType(t);

            ts.checkCycles(t.toReference());
        }
        else if (ts.Object() != this.type && !ts.Object().fullName().equals(this.type.fullName())) {
            // the supertype was not specified, and the type is not the same
            // as ts.Object() (which is typically java.lang.Object)
            // As such, the default supertype is ts.Object().
            this.type.superType(ts.Object());
        }
        else {
            // the type is the same as ts.Object(), so it has no supertype.
            this.type.superType(null);
        }    
    }

    protected String theName() {
        return ((ExtensionType)type()).extensionName();
    }

    protected String commaDelimited(String keyword, List l) {
        if (l == null || l.size() == 0)
            return "";
        Iterator it = l.iterator();
        String r = keyword + " " + it.next();
        while (it.hasNext())
            r += ", " + it.next();
        return r;
    }

    public void prettyPrint(CodeWriter w, PrettyPrinter pp) {
        if (!isTop && !isLocalTop)
            return;

        String theName = theName();

//      System.out.println("expander " + theName + " " + (isLocalTop ? "(local) " :
//      "") + ((ExtensionType)type()).cases() + (isLocalTop ? "(parent is " +
//      ((ExtensionType)type()).parent() + ")" : ""));

        w.write(flags().translate() + "class " + (isLocalTop ? position().file().split("\\.")[0] + "$" : "") + theName);
        if (isLocalTop)
            w.write(" extends " + theName);
        w.write(" {");
        w.newline(4);
        w.begin(0);

        if (isTop) {
            w.write("public static int idCount = 0;");
            w.newline(0);
        }
        w.write("public static int id;");
        w.newline(0);
        w.write("static { id = idCount++; }");
        w.newline(0);

        if (!isLocalTop) {
            w.write("public static final java.util.WeakHashMap<" + forClass() + ", " + theName + "$$tate> cache = new java.util.WeakHashMap<" + forClass() + ", " + theName + "$$tate>();");
            w.newline(0);
        }
        String baseType = (isLocalTop ? ((ExtensionType)type()).parent() : type()).superType().toString();
        String baseWrapperType = theName + "$" + mangle(baseType);
        boolean hasState = printState(w, pp) > 0;
        Vector sorted = new Vector();
        for (Iterator i = cases.iterator(); i.hasNext(); ) {
            ExtensionDecl_c ed = (ExtensionDecl_c)i.next();
            Type bt = ed.forClass().type(), st = null;
            boolean inserted = false;
            for (int idx = 0; idx < sorted.size(); idx += 1) {
                ExtensionDecl_c ed2 = (ExtensionDecl_c)sorted.elementAt(idx);
                Type bt2 = ed2.forClass().type();
                if (bt.isSubtype(bt2)) {
                    sorted.insertElementAt(ed, idx);
                    inserted = true;
                    break;
                }
            }
            if (!inserted)
                sorted.add(ed);

            for (Iterator j = cases.iterator(); j.hasNext(); ) {
                Type t = ((ExtensionDecl_c)j.next()).forClass().type();
                if (bt.descendsFrom(t) && (st == null || t.descendsFrom(st)))
                    st = t;
            }
            if (st == null) {
                ExtensionType parent = ((ExtensionType)type()).parent();
                if (parent != null) {
//                  System.out.println("parent is "+parent + ", with cases "+ parent.cases());
//                  System.out.println("bt = " + bt);
                    for (Iterator j = parent.cases().iterator(); j.hasNext(); ) {
                        Type t = (Type)j.next();
//                      System.out.println("t = "+ t);
                        if (t.descendsFrom(bt)) {
                            System.err.println("error: local refinements can only exist for leaves of original type hierarchy");
                            System.exit(1);
                        }
                        if (bt.descendsFrom(t) && (st == null || t.descendsFrom(st)))
                            st = t;
                    }
                    if (st == null) {
                        System.err.println("error: cannot have a local expander refinement for a supertype of the original expander's base type");
                        System.exit(1);
                    }
                    // System.out.println("st = "+st);
                }
                else
                    st = forClass().type();
            }

            w.write("public static class " + theName + "$" + mangle(bt) + " ");
            if (bt == forClass().type() && ((ExtensionType)type()).parent() == null)
                w.write(commaDelimited("implements", interfaces));
            else
                w.write("extends " + theName + "$" + mangle(st));
            w.write(" {");
            w.newline(4);
            w.begin(0);
            if (bt == forClass().type() && ((ExtensionType)type()).parent() == null) {
                w.write("protected final " + forClass() + " instance;");
                w.newline(0);
                if (hasState) {
                    w.write("protected final " + theName + "$$tate state;");
                    w.newline(0);
                }
            }
            w.write("protected " + theName + "$" + mangle(bt) + "(final " + baseType + " instance" + (hasState ? (", " + theName + "$$tate state") : "") + ") {");
            w.newline(4);
            w.begin(0);
//          System.out.println("bt = "+bt + ", forclass = "+forClass().type());
            if (bt == forClass().type() && !isLocalTop) {
                w.write("this.instance = instance;");
                w.newline(0);
                if (hasState) {
                    w.write("this.state = state;");
                    w.newline(0);
                }
            }
            else
                w.write("super(instance" + (hasState ? ", state" : "") + ");");
            w.end();
            w.newline(0);
            w.write("}");
            w.newline(0);
            w.write("public int $creator$() { return id; }");
            w.newline(0);
            if (bt == forClass().type())
                implementRequiredForwardingMethods(w);
            ed.printMethods(w, pp);
            w.end();
            w.newline(0);
            w.write("}");
            w.newline(0);
        }
//      String baseWrapperType = theName + "$" + mangle((isLocalTop ?
//      ((ExtensionType)type()).parent() : type()).superType());
        w.write("public static synchronized " + baseWrapperType + " expand(final " + topForClass + " instance) {");
        w.newline(4);
        w.begin(0);
        /*
         * Sunday - this will only work if baseWrapperType is X$Object. Should I put it
         * in? Probably not, seems dangerous - could end up not throwing away existing
         * wrapper that was produced by another version of the same expander (a local
         * refinement that's out of scope).
         * 
         * w.write("if (instance instanceof " + baseWrapperType + ") return (" +
         * baseWrapperType + ")instance;"); w.newline(0);
         */
        w.write(baseWrapperType + " r = null;");
        w.newline(0);
        w.write(theName + "$$tate state = cache.get(instance);");
        w.newline(0);
        w.write("if (state == null) {");
        w.newline(4);
        w.begin(0);
        w.write("state = new " + theName + "$$tate();");
        w.newline(0);
        w.write("state.instance$ = new java.lang.ref.WeakReference<" + topForClass + ">(instance);");
        w.newline(0);
        w.write("cache.put(instance, state);");
        w.end();
        w.newline(0);
        w.write("}");
        w.newline(0);
        w.write("else if (state.wrapper$ != null) {");
        w.newline(4);
        w.begin(0);
        w.write("r = state.wrapper$.get();");
        w.newline(0);
        w.write("if (r != null && r.$creator$() == id)");
        w.newline(4);
        w.begin(0);
        w.write("return r;");
        w.end();
        w.end();
        w.newline(0);
        w.write("}");
        w.newline(0);
        boolean first = true;
        for (Iterator i = sorted.iterator(); i.hasNext(); ) {
            ExtensionDecl_c ed = (ExtensionDecl_c)i.next();
            if (first)
                first = false;
            else
                w.write("else ");
            if (i.hasNext() || ((ExtensionType)type()).parent() != null) {
                Type t = ed.forClass().type();
                t = ((EJavaTypeSystem)t.typeSystem()).rawify(t);
                w.write("if (instance instanceof " + t + ") ");
            }
            w.write("r = new " + theName + "$" + mangle(ed.forClass()) + "(" + /*
             * (" +
             * ed.forClass() + ")
             */ "instance" + (hasState ? ", state" : "") + ");");
            w.newline(0);
        }
        if (((ExtensionType)type()).parent() != null) {
            w.write("else return " + theName + ".expand(instance);");
            w.newline(0);
        }
        w.write("state.wrapper$ = new java.lang.ref.WeakReference<" + baseWrapperType + ">(r);");
        w.newline(0);
        w.write("return r;");
        w.end();
        w.newline(0);
        w.write("}");
        w.end();
        w.newline(0);
        w.write("}");
        w.newline(0);
    }

    public void printMethods(CodeWriter w, PrettyPrinter pp) {
        for (Iterator i = body().members().iterator(); i.hasNext(); ) {
            ClassMember cm = (ClassMember)i.next();
            if (!(cm instanceof MethodDecl))
                continue;
            cm.prettyPrint(w, pp);
            w.newline(0);
        }
    }

    protected int printState(CodeWriter w, PrettyPrinter pp) {
        int numFields = 0;
        if (!isLocalTop) {
            w.write("public static class " + theName() + "$$tate {");
            w.newline(4);
            w.begin(0);
            w.write("public java.lang.ref.WeakReference<" + topForClass + "> instance$ = null;");
            w.newline(0);
            w.write("public java.lang.ref.WeakReference<" + theName() + "$" + mangle(topForClass) + "> wrapper$ = null;");
            w.newline(0);
        }
        for (Iterator i = body().members().iterator(); i.hasNext(); ) {
            ClassMember cm = (ClassMember)i.next();
            if (!(cm instanceof FieldDecl))
                continue;
            numFields += 1;
            if (!isLocalTop) {
                cm.prettyPrint(w, pp);
                w.newline(0);
            }
        }
        if (!isLocalTop) {
            w.end();
            w.newline(0);
            w.write("}");
            w.newline(0);
        }
        return numFields;
    }

    protected boolean implementRequiredForwardingMethods(CodeWriter w) {
        boolean r = false;
        Iterator interfaces = ((ClassType)type()).interfaces().iterator();
        while (interfaces.hasNext()) {
            ClassType ifc = (ClassType)interfaces.next();
            Iterator methods = ifc.methods().iterator();
            while (methods.hasNext()) {
                MethodInstance method = (MethodInstance)methods.next();
                if (!((ClassType)type()).hasMethod(method)) {
                    w.write(method.flags().clearAbstract().translate());
                    w.write(method.returnType().toString());
                    w.write(" "+method.name()+"(");
                    int n = 0;
                    Iterator argTypes = method.formalTypes().iterator();
                    while (argTypes.hasNext()) {
                        Type t = (Type) argTypes.next();
                        n++;
                        w.write(t.toString()+" a"+n);
                        if (argTypes.hasNext())
                            w.write(", ");
                    }
                    w.write(") ");
                    if (!method.throwTypes().isEmpty()) {
                        w.write("throws ");
                        for (Iterator i = method.throwTypes().iterator(); i.hasNext(); ) {
                            TypeNode tn = (TypeNode) i.next();
                            w.write(tn.toString());
                            if (i.hasNext())
                                w.write(", ");
                            else
                                w.write(" ");
                        }
                    }
                    w.write("{");
                    w.newline(4);
                    w.begin(0);
                    if (!method.returnType().toString().equals("void"))
                        w.write("return ");
                    w.write("instance."+method.name()+"(");
                    for (int an = 1; an <= n; an++) {
                        w.write("a"+an);
                        if (an < n)
                            w.write(", ");
                    }
                    w.write(");");
                    w.end();
                    w.newline(0);
                    w.write("}");
                    w.newline(0);
                    r = true;
                }
            }
        }
        return r;
    }

    public Context enterScope(Node child, Context c) {
        /*
         * // This is what the one from ClassDecl_c does: if (child == this.body) {
         * TypeSystem ts = c.typeSystem(); c = c.pushClass(type,
         * ts.staticTarget(type).toClass()); } return super.enterScope(child, c);
         */
        if (child == this.body && isTop) {

//          THIS IS NEW XXXXXX Jan 23, 2006: TRYING IT OUT!
//          c = c.pushClass(type(), c.typeSystem().staticTarget(type).toClass());
//          END OF NEW EXPERIMENTAL STUFF

            ((EJavaContext)c).addExtension((ExtensionType)type());
        }
        return super.enterScope(child, c);
    }

    // give this ast node the type ExtensionType so we can tell it apart from
    // class declarations
    public NodeVisitor buildTypesEnter(TypeBuilder tb) throws SemanticException {
        tb = ((EJavaTypeBuilder)tb).pushExtension(position(), flags, name, extensionName, isTop);
        ParsedClassType ct = tb.currentClass();
        // Member extensions of interfaces are implicitly public and static.
        if (ct.isMember() && ct.outer().flags().isInterface())
            ct.flags(ct.flags().Public().Static());
        return tb;
    }

    protected boolean defaultConstructorNeeded() {
        return false;
    }

    public Node typeCheck(TypeChecker tc) throws SemanticException {
        if (!isTop()) {
            if (flags != Flags.NONE)
                throw new SemanticException("only the top extension may have modifiers");
            if (((ClassType)superClass().type()).flags().isInterface())
                throw new SemanticException("only the top extension may extend an interface type");
            ExtensionType et = null;
            try {
                et = (ExtensionType)tc.context().importTable().find(extensionName());
            }
            catch (SemanticException e) {
                throw new SemanticException("extension type "+extensionName()+" not found");
            }
            catch (ClassCastException e) {
                throw new SemanticException(extensionName()+" is not an extension type");
            }
            ExtensionDecl_c r = (ExtensionDecl_c)commonTypeCheck(tc);
            r.topForClass = ((ExtensionType)tc.context().find(((ExtensionType)r.type()).extensionName())).superType();
            ((ExtensionType)r.type()).parent((ExtensionType)tc.context().find(((ExtensionType)r.type()).extensionName()));
            return r;
        }
        else {
            ExtensionDecl_c r = (ExtensionDecl_c)commonTypeCheck(tc);
            r.topForClass = r.forClass().type();
            return r;
        }
    }


    protected Node commonTypeCheck(TypeChecker tc) throws SemanticException {
        if (this.type().isNested() && (this.type() instanceof Named)) {
            // The class cannot have the same simple name as any enclosing class.
            ClassType container = this.type.outer();

            while (container instanceof Named) {
                if (!container.isAnonymous()) {
                    String name = ((Named) container).name();

                    if (name.equals(extensionName))
                        throw new SemanticException("Cannot declare member " + "class \"" + this.type + "\" inside class with the " + "same name.", position());
                }
                if (container.isNested())
                    container = container.outer();
                else
                    break;
            }
            if (this.type().isLocal()) {
                // a local class name cannot be redeclared within the same
                // method, constructor or initializer, and within its scope
                Context ctxt = tc.context();

                if (ctxt.isLocal(extensionName)) {
                    // something with the same name was declared locally.
                    // (but not in an enclosing class)
                    Named nm = ctxt.find(extensionName);
                    if (nm instanceof Type) {
                        Type another = (Type)nm;
                        if (another.isClass() && another.toClass().isLocal())
                            throw new SemanticException("Cannot declare local " + "class \"" + this.type + "\" within the same " + "method, constructor or initializer as another " + "local class of the same name.", position());
                    }
                }
            }
        }

        // check that inner classes do not declare member interfaces
        if (type().isMember() && flags().isInterface() &&
                type().outer().isInnerClass())
            // it's a member interface in an inner class.
            throw new SemanticException("Inner classes cannot declare " + "member interfaces.", this.position());             
        // Make sure that static members are not declared inside inner classes
        if (type().isMember() && type().flags().isStatic() && type().outer().isInnerClass())
            throw new SemanticException("Inner classes cannot declare static " + "member classes.", position());
        /*
         * Alex: it's ok to have "interface extensions", so skip this check.
         */
        if (type.superType() != null)
            if (! type.superType().isClass())
                throw new SemanticException("Cannot extend non-class \"" + type.superType() + "\".", position());

        /*
         * Alex: it's ok to extend final classes, so skip this check. if
         * (type.superType().toClass().flags().isFinal()) throw new
         * SemanticException("Cannot extend final class \"" + type.superType() + "\".",
         * position());
         */

        TypeSystem ts = tc.typeSystem();

        try {
            if (type.isTopLevel())
                ts.checkTopLevelClassFlags(type.flags());
            if (type.isMember())
                ts.checkMemberClassFlags(type.flags());
            if (type.isLocal())
                ts.checkLocalClassFlags(type.flags());
        }
        catch (SemanticException e) {
            throw new SemanticException(e.getMessage(), position());
        }

        // Alex: we need to handle conformance checking slightly differently
        // ts.checkClassConformance(type);
        if (isTop) {
            for (Iterator i = interfaces().iterator(); i.hasNext(); )
                ensureInterfaceIsImplemented((ClassType)((TypeNode)i.next()).type(), getAllClassesAndInterfaces(), (EJavaTypeSystem)ts);
        }

        return this;
    }

    protected List getAllClassesAndInterfaces() {
        List r = getAllClassesAndInterfacesHelper((ClassType)type().superType());
        r.add(type());
        return r;
    }

    protected List getAllClassesAndInterfacesHelper(ClassType c) {
        List r = new LinkedList();
        r.add(c);
        ClassType base = (ClassType)c.superType();
        if (base != null)
            r.addAll(getAllClassesAndInterfacesHelper(base));
        for (Iterator is = c.interfaces().iterator(); is.hasNext(); ) {
            ClassType i = (ClassType)is.next();
            r.addAll(getAllClassesAndInterfacesHelper(i));
        }  
        return r;
    }

    protected void ensureInterfaceIsImplemented(ClassType ifc, List allSources, EJavaTypeSystem ts) throws SemanticException {
        // Make sure that someone ("for" class/interface + all of its interfaces)
        // implemented this interface.
        for (Iterator methods = ifc.methods().iterator(); methods.hasNext(); ) {
            MethodInstance method = (MethodInstance)methods.next();
//          System.out.println("looking for "+method);
            boolean found = false;
            for (Iterator sources = allSources.iterator(); sources.hasNext(); ) {
                ClassType source = (ClassType)sources.next();
//              System.out.println("in "+source);
//              System.out.println(source.hasMethod(method)+", "+ts.isAccessible(method,
//              source));
                if (source.hasMethod(method)) { // && ts.isAccessible(method, source)) {
//                  System.out.println("found it!");
                    found = true;
                    break;
                }
            }
            if (!found)
                throw new SemanticException(method+" must be implemented "+ "in extension "+extensionName(), position());
        }
        // Make sure that the interfaces implemented by this interface are also
        // implemented
        for (Iterator i = ifc.interfaces().iterator(); i.hasNext(); )
            ensureInterfaceIsImplemented((ClassType)i.next(), allSources, ts);
    }
}

