package polyglot.ext.ejava.visit;

import polyglot.ast.NodeFactory;
import polyglot.ext.ejava.types.EJavaTypeSystem;
import polyglot.ext.ejava.types.ExtensionType;
import polyglot.ext.jl5.visit.JL5TypeBuilder;
import polyglot.frontend.Job;
import polyglot.types.*;
import polyglot.util.Position;
import polyglot.visit.TypeBuilder;

public class EJavaTypeBuilder extends JL5TypeBuilder {

    public EJavaTypeBuilder(Job job, TypeSystem ts, NodeFactory nf) {
        super(job, ts, nf);
    }

    public TypeBuilder pushExtension(Position pos, Flags flags, String name, String extensionName,
            boolean isTop) throws SemanticException {
        ParsedClassType t = newExtension(pos, flags, name, extensionName, isTop);
        return pushClass(t);
    }

    protected ParsedClassType newExtension(Position pos, Flags flags, String name,
            String extensionName, boolean isTop) throws SemanticException {
        EJavaTypeSystem ts = (EJavaTypeSystem) typeSystem();
        ExtensionType ct = ts.createExtensionType(ts.defaultClassInitializer(), this.job.source());
        ct.isTop(isTop);
        ct.extensionName(extensionName);
        // the rest of this method is exactly like TypeBuilder.newClass()
        if (inCode) {
            ct.kind(ClassType.LOCAL);
            ct.outer(currentClass());
            ct.flags(flags);
            ct.name(name);
            ct.position(pos);

            if (currentPackage() != null) {
                ct.package_(currentPackage());
            }

            return ct;
        } else if (currentClass() != null) {
            ct.kind(ClassType.MEMBER);
            ct.outer(currentClass());
            ct.flags(flags);
            ct.name(name);
            ct.position(pos);

            currentClass().addMemberClass(ct);

            if (currentPackage() != null) {
                ct.package_(currentPackage());
            }

            // if all the containing classes for this class are member
            // classes or top level classes, then add this class to the
            // parsed resolver.
            ClassType container = ct.outer();
            boolean allMembers = (container.isMember() || container.isTopLevel());
            while (container.isMember()) {
                container = container.outer();
                allMembers = allMembers && (container.isMember() || container.isTopLevel());
            }

            if (allMembers) {
                typeSystem().parsedResolver()
                        .addNamed(typeSystem().getTransformedClassName(ct), ct);
            }

            return ct;
        } else {
            ct.kind(ClassType.TOP_LEVEL);
            ct.flags(flags);
            ct.name(name);
            ct.position(pos);

            if (currentPackage() != null) {
                ct.package_(currentPackage());
            }

            Named dup = ((CachingResolver) typeSystem().systemResolver()).check(ct.fullName());

            if (dup != null && dup.fullName().equals(ct.fullName()) && ct.isTop()) {
                throw new SemanticException("Duplicate class \"" + ct.fullName() + "\".", pos);
            }

            // if (ct.isTop()) {
            typeSystem().parsedResolver().addNamed(ct.fullName(), ct);
            ((CachingResolver) typeSystem().systemResolver()).addNamed(ct.fullName(), ct);
            // }

            return ct;
        }
    }

}
