package polyglot.ext.ejava.visit;

import polyglot.ast.*;
import polyglot.ext.ejava.ast.EJavaNodeFactory;
import polyglot.ext.ejava.types.ExtensionType;
import polyglot.frontend.Job;
import polyglot.types.SemanticException;
import polyglot.types.Type;
import polyglot.types.TypeSystem;
import polyglot.visit.AscriptionVisitor;

/*******************************************************************************
 * Visitor that transforms expressions in the extension's body in order to make
 * code generation work.
 ******************************************************************************/
public class ExtensionBodyTransformer extends AscriptionVisitor {
    public ExtensionBodyTransformer(Job job, TypeSystem ts, NodeFactory nf) {
        super(job, ts, nf);
    }

    protected ExtensionType extension() {
        Type t = context().currentClass();
        if (t instanceof ExtensionType)
            return (ExtensionType) t;
        else
            return null;
    }

    protected boolean insideExtension() {
        return extension() != null;
    }

    // transform "this" and "super" to either "this" or "instance"
    public Expr ascribe(Expr e, Type toType) throws SemanticException {
        if (e instanceof Special) {
            Special es = (Special) e;
            if (insideExtension()) {
                ExtensionType et = extension();
                Type extensionToType = et, extensionFromType = et.superType();
                if (es.kind().equals(Special.THIS)) {
                    if (extensionFromType.isImplicitCastValid(toType) || toType.equals(ts.String()))
                        return nf.Cast(es.position(),nf.CanonicalTypeNode(es.position(), extensionFromType),nf.Special(es.position(), 
                                EJavaNodeFactory.UNEXTENDED).type(extensionFromType)).type(extensionFromType);
                    else
                        return es.type(et);
                }
                // it's not "this", so it's "super"
                else if (extension().isTop())
                    throw new SemanticException("the top extension cannot use super");
                else
                    return es;
            } else
                return es;
        } else if (e instanceof Field) {
            if (insideExtension())
                return ((Field)e).targetImplicit(false);
            return e;
        } else if (e instanceof Call) {
            if (insideExtension())
                return ((Call)e).targetImplicit(false);
            return e;
        }
        return super.ascribe(e, toType);
    }
}
