package polyglot.ext.ejava.visit;

import java.util.HashMap;

import polyglot.ast.Node;
import polyglot.ast.NodeFactory;
import polyglot.ext.ejava.ast.ExtensionDecl;
import polyglot.ext.ejava.types.ExtensionType;
import polyglot.frontend.Job;
import polyglot.types.SemanticException;
import polyglot.types.TypeSystem;
import polyglot.visit.ContextVisitor;

/*******************************************************************************
 * Visitor that adds all of the cases of an extension to the top-level
 * extension. This guy side-effects ExtensionDecl ast nodes.
 ******************************************************************************/
public class ExtensionConsolidator extends ContextVisitor {
    protected HashMap extensions;

    public ExtensionConsolidator(Job job, TypeSystem ts, NodeFactory nf, HashMap extensions) {
        super(job, ts, nf);
        this.extensions = extensions;
    }

    protected Node leaveCall(Node n) throws SemanticException {
        if (n instanceof ExtensionDecl) {
            ExtensionDecl nn = (ExtensionDecl) n;

            ExtensionDecl r;
            String compilationUnit = nn.position().file();
            if (context.package_() != null)
                compilationUnit = context.package_() + "." + compilationUnit;
            HashMap extensionsInThisCU = (HashMap) extensions.get(compilationUnit);
            if (extensionsInThisCU == null) {
                extensionsInThisCU = new HashMap();
                extensions.put(compilationUnit, extensionsInThisCU);
            }
            if (!nn.isTop()) {
                String name = ((ExtensionType) nn.type()).extensionName();
                ExtensionDecl ed = (ExtensionDecl) extensionsInThisCU.get(name);
                if (ed == null) {
                    nn.promoteToLocalTop();
                    // System.out.println("in cu "+compilationUnit+", promoted
                    // to local top level: "+n);
                    extensionsInThisCU.put(name, n);
                    ed = nn;
                }
                ((ExtensionType) ed.type()).addCase(nn.type().superType());
                ed.addCase(nn);
                r = nn;
            } else {
                r = (ExtensionDecl) extensionsInThisCU.get(nn.name());
                ((ExtensionType) r.type()).addCase(r.type().superType());
                r.addCase(r);
            }
            return r;
        }
        return super.leaveCall(n);
    }
}
